%**********************************************************************
%
% (C) Copyright 1989
% File:             TXT2HLP.RED
% Object File:      TXT2HLP
%
% Convert a list of .txt and .red files into a single .HLP file. This
% involves filling in all labels with their byte addresses in the
% final file. Assumes that end of line is two characters: 13 then 10.
% Assumes that all input is ASCII. The intercepted commands are @node,
% @related, and @alternate.
% 
% The system is divided into *topics* of help (usually by package) and
% *items* within each class.
%
% UO-LISPisms:
%  COMPRESS - must be able to generate list (compress '(!( a !) )) = (A)
%  N2I - Convert number to an id.
%  LINELENGTH - With 0 argument means no auto-LF's
%  SORT - Sorts a list (first arg) based on 2 arg function which is 
%            second argument.
%
% Revision History: (Created 10/2/89)
%   11/9/89 Comment out explode2n for UO-LISP.
%   11/11/89 Convert to RLISP.
%   12/12/89 Switch to list/type format. Don't sort any more. Add the
%       @setclass command.
%   12/1/91 Fix bugs, comments.
%   12/21/91 Fix some of the UO-LISPism's to make more portable.
%   12/24/91 Remove dependency on 26 being EOF. Use h!:syseof. Comment
%     some stuff, fix bugs.
%   12/29/91 Formalize topic/item relationships.
%
%***********************************************************************

GLOBAL '(CPTR!*                          % File byte pointer.
         NCNT!*                          % Number of names encountered.
         CURTOPICNAME!*                  % Current topic name.
         CURTOPICLIST!*                  % List associated with this name.
         TOPICS!*                        % List of topics by name.
         NODECNT!*                       % No. of bytes in @node label
         RELATEDCNT!*                    % No. of bytes in @related label.
         CNODE!*                         % Current node file pointer.
         ITEMS!*                         % A list of referenced items.
        );

NODECNT!* := 28;                         % No. of bytes in @node
RELATEDCNT!* := 10;                      % No. of bytes in @related.

GLOBAL '(!*NOISY);                       % When T, display what's happening.
SWITCH NOISY;
!*NOISY := T;

% System EOF. This is also used in the base system. This needs be set by
% each system (UNIX uses 4, MS-DOS uses 26, others may be worse).
GLOBAL '(H!:SYSEOF H!:EOL);
H!:SYSEOF := 26;
H!:EOL := 10;


symbolic procedure txt2hlp(fins,fout1,fout2); 
% Convert the list of files FINS into a help file FOUT1 and index FOUT2.
begin scalar fouth, xx;
%  linelength 0;                          % UO-LISP specific - no CR's
  % 1) Clear things out, search all the files for topics and items.
  CPTR!* := 0; 
  NCNT!* := 0; 
  CURTOPICNAME!* := CURTOPICLIST!* := TOPICS!* := NIL; 
  for each f in fins
      do <<prin2 "  "; prin2 f; prin2 "..."; findnodes f; terpri() >>; 

  % 2) Add the last topic to the list or replace its current entry.
  if xx := assoc(CURTOPICNAME!*, TOPICS!*) then cdr xx := CURTOPICLIST!*
     else TOPICS!* := (CURTOPICNAME!* . CURTOPICLIST!*) . TOPICS!*;
  TOPICS!* := reversip TOPICS!*;

  % 3) Make some noise about what happened.
  prin2 "Finished with files. Final size will be ";  prin1 CPTR!*; 
    prin2 ", nodes = "; print NCNT!*; 
  prin2 "The following "; prin2 length TOPICS!*; prin2t " topics are defined:";
    for each x in TOPICS!*
        do << prin2 "  "; prin1 car x; prin2 " [";
              prin2 length cdr x;
              prin2t " entries]" >>;

  % 4) Now check cross references and print out any undefined references.
  xx := nil;
  for each x in ITEMS!* 
      do if not anyhelp x then
           << if not xx then xx := prin2t "The following are undefined:";
              print x >>;

  % 5) Merge the new input file into the master help file.
  prin2 "Writing master file ";  prin2t fout1; 
  fouth := wrs open(fout1,'output); 
  for each x in fins do copynodes x;
  close wrs fouth; 

  % 6) Now write the index file.
  prin2 "Writing index file "; prin2 fout2; prin2t "...";
  fouth := wrs open(fout2,'output); 
  prin2 "(";
  for each x in TOPICS!*
      do << prin2 "( "; print car x;
            for each elt in cdr x do print elt;
            prin2t ")" >>;
  prin2t ")";
  close wrs fouth; 
  prin2t "Done!";
  return NIL;
end;

% Look for items:
%  Fully qualified:  (topic . item)
%  local:            item  (also look under MASTER if not local).
expr procedure anyhelp x;
  if atom x then
     if assoc(x, CURTOPICLIST!*) then t else anyhelp('MASTER . x)
  else (if w then assoc(cdr x, cdr w)) where w :=  assoc(car x, TOPICS!*);

symbolic procedure findnodes f; 
% Find all the @node, @related, @alternate and @setclass commands in file f.
% Compute how many bytes they will use so pass2 can fill in the right byte
% count.
begin scalar finh;
  finh := rds open(f,'input);
  findnodes1();
  close rds finh
end;


symbolic procedure findnodes1; 
% Find @node, @replace, @alternate, and @setclass functions. Compute the
% number of bytes each will use. 
begin scalar c, l, le, comm, xx;
  % 1) Skip characters to an @ counting as we go along.
loop: if (c := readchar()) eq h!:syseof then return t; 
  if c eq 64 then go to s1; 
  CPTR!* := add1 CPTR!*; 
  go to loop; 
  % 2) Read characters to end of @xxx token.
s1: l := le := list c;
loop2: if (c := readchar()) eq h!:syseof then return t
   else if not(c <= 32 or c eq 123 or c eq 125) then
      << rplacd(le, list n2i c); le := cdr le; go to loop2 >>;
  % 3) If it's not one we're interested in, count and skip it.
  comm := xcompress cdr l;
  if not memq(comm, '(node related alternate setclass)) then
     << CPTR!* := CPTR!* + length l + 1;
        go to loop >>;
  % 4) Get the rest of the line and build into a list.
  l := le := list '!(;
loop3: if (c := readchar()) eq H!:EOL then go to fini
         else if c eq h!:syseof then error(0, list("@", comm, "at eof"));
      rplacd(le, list n2i c);
      le := cdr le;
      go to loop3;
  % 5) Build a list of the argumentss
fini: rplacd(le, list '!));
  l := xcompress l;
 
 % 6) @node a b c d -> Build a dummy class if necessary, warn if already
  %   a node, list the topic. 
  if comm eq 'node then 
    << if not CURTOPICNAME!* and null CURTOPICLIST!* then
          << prin2t "*** No current class, creating MASTER";
             CURTOPICNAME!* := 'MASTER;
             CURTOPICLIST!* := NIL >>;
       if assoc(car l, CURTOPICLIST!*) then
          << prin2 "*** "; prin1 car l; prin2 " is already on list ";
             print CURTOPICNAME!* >>
       else CURTOPICLIST!* := (car l . (CNODE!* := CPTR!*)) . CURTOPICLIST!*;
       if !*NOISY then << prin1 car l; prin2 ": "; print CPTR!* >>;
       CPTR!* := CPTR!* + NODECNT!*;
       for each x in l
           do << if atom x then x := CURTOPICNAME!* . x;
                 if not x member ITEMS!* then ITEMS!* := x . ITEMS!* >>;
       NCNT!* := add1 NCNT!* >>

  % 7) @related a b c ... -> Add these to the list of names defined.
  else if comm eq 'related then
    << CPTR!* := CPTR!* + RELATEDCNT!* + 7 * length l;
       for each x in l
           do << if atom x then x := CURTOPICNAME!* . x;
                 if not x member ITEMS!* then ITEMS!* := x . ITEMS!* >> >>

  % 8) If we have an alternate name, add to the list.
  else if comm eq 'alternate then
    << if assoc(car l, CURTOPICLIST!*) then
          << prin2 "*** "; prin1 car l; prin2 " is already on list ";
             print CURTOPICNAME!* >>
         else CURTOPICLIST!* := (car l . CNODE!*) . CURTOPICLIST!*;
       xx := if atom car l then CURTOPICNAME!* . car l else car l;
       if not (xx member ITEMS!*) then ITEMS!* := xx . ITEMS!*;
       NCNT!* := add1 NCNT!* >>

  % 9) Here, we're changing topics. Save the old one first.
  else if comm eq 'setclass then
    << if CURTOPICNAME!* then
         if xx := assoc(CURTOPICNAME!*, TOPICS!*) then
            rplacd(xx, CURTOPICLIST!*)
         else TOPICS!* := (CURTOPICNAME!* . CURTOPICLIST!*) . TOPICS!*;
       if !*NOISY then
         << prin2 ">>> Switch to class: "; print car l >>;
       CURTOPICNAME!* := car l;
       CURTOPICLIST!* :=
         if xx := assoc(CURTOPICNAME!*, TOPICS!*) then cdr xx >>;
  go to loop
end;

symbolic procedure copynodes f; 
% Now we copy the input file to the output file changing names to file
% pointers.
begin scalar finh;
  finh := rds open(f,'input);
  copynodes1();
  close rds finh
end;


symbolic procedure copynodes1; 
% Copy a file replacing @node, @related and friends.
begin scalar c, l, le, comm;
  % 1) Read to an @ command.
loop: if (c := readchar()) eq h!:syseof then return t; 
      if c eq 64 then go to s1; 
      writechar c; 
      go to loop; 

  % 2) Read an @ command.
s1: l := le := list n2i c;
loop2: if (c := readchar()) eq h!:syseof then return t
   else if not(c <= 32 or c eq 123 or c eq 125) then
      << cdr le := list n2i c; le := cdr le; go to loop2 >>;
   comm := xcompress cdr l;

   % 3) Write out one of the non-selected commands.
   if not memq(comm, '(node related alternate setclass)) then
      << CPTR!* := CPTR!* + length l + 1;
         for each x in l do writechar i2n x;
         writechar c;
         go to loop >>;

   % 4) Build the arguments following the command.
   l := le := list '!(;
loop3: if (c := readchar()) eq 10 then go to fini
         else if c eq h!:syseof then error(0, list("@", comm, "at eof"));
      cdr le := {n2i c};
      le := cdr le;
      go to loop3;
fini: cdr le := {'!)};
    l := xcompress l;

    % 5) If we have an @node, locate it and print its 3 prongs.
    if comm eq 'node then 
       << mapc('(64 110 111 100 101), function writechar); 
          p6dig wherehelp cadr l;
          p6dig wherehelp caddr l;
          p6dig wherehelp cadddr l;
          terpri(); 
          go to loop >>

    % 6) Write out a related group.
    else if comm eq 'related then
       << mapc('(64 114 101 108 97 116 101 100), function writechar);
          for each x in l do p6dig wherehelp x;
          terpri();
          go to loop >>

    % 7) If we have setclass, then set the current class and switch lists.
    %    The command is not written to the destination file.
    else if comm eq 'setclass then
        << CURTOPICNAME!* := car l;
           CURTOPICLIST!* := cdr assoc(CURTOPICNAME!*, TOPICS!*);
           go to loop >>
    else go to loop
end;

expr procedure wherehelp x;
% Where is the help on item x (see anyhelp earlier). 
if atom x then wherehelp1 x else wherehelp2(car x, cdr x);

expr procedure wherehelp1 x;
(if w then cdr w else wherehelp2('MASTER, x))
   where w := assoc(x, CURTOPICLIST!*);

expr procedure wherehelp2(tp, it);
begin scalar loc;
  if null(loc := assoc(tp, TOPICS!*)) then return nil;
  if null(loc := assoc(it, cdr loc)) then return nil;
  return cdr loc
end;

symbolic procedure p6dig n; 
    if numberp n
      then begin scalar d,i; 
              writechar 32; 
              i := 1; 
            loop: 
              if igreaterp(i,6) then go to done; 
              d := (48+remainder(n,10)).d; 
              n := n/10; 
              i := add1 i;
              go to loop; 
            done: 
              for each x in d do writechar x
           end
     else prin2 " XXXXXX";

% Not part of St. Lisp, but is part of UO-LISP.
%symbolic procedure explode2n a; mapcar(explode2 a,function i2n);


expr procedure xcompress l;
% This is a compress that interns ids. It's necessary for real St. Lisp.
 if idp(l := compress l) then intern l else internl l;

expr procedure internl l;
if idp l then intern l
  else if atom l then l
  else internl car l . internl cdr l;

end;

