%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:FUNCTION-PRIMITIVES.SL 
% Title:        Primitives used by PUTD/GETD and EVAL/APPLY 
% Author:       Eric Benson, Steve Lowder 
% Created:      23 August 1981 
% Modified:     29-Aug-84 10:08:36 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Compiletime:  PL:SYS-CONSTS.B 
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Every ID has a "function cell".  It does not necessarily contain a legal
% Lisp item, and therefore should not be accessed directly by Lisp functions.
% In this implementation the function cell contains an instruction to be
% executed.  There are 3 possibilites for this instruction, for which the
% following predicates and updating functions exist:
%
%    FUnBoundP(ID)          -- the function is not defined
%    FLambdaLinkP(ID)       -- the function is interpreted
%    FCodeP(ID)             -- the function is compiled
%
%    MakeFUnBound(ID)       -- undefine the function
%    MakeFLambdaLink(ID)    -- specify that the function is interpreted
%    MakeFCode(ID, CodePtr) -- specify that the function is compiled,
%                              and that the code resides at the address
%                              associated with CodePtr
%
%    GetFCodePointer(ID)    -- returns the contents of the function cell as a
%                              code pointer.
%
% These functions currently check that they have proper arguments, but this may
% change since they are only used by functions that have checked them already.
%

(compiletime (load sys-consts))

(fluid '(lambda-link-address undefined-function-address))

(setq lambda-link-address
  (getmem (wplus2 symfnc (wtimes2 (id2int 'compiledcallinginterpreted) 4))))

(setq undefined-function-address
      (getmem (wplus2 symfnc (wtimes2 (id2int 'undefinedfunction) 4))))

(de funboundp (u)
  % does U not have a function defn?
  (if (idp u)
    (weq (symfnc (idinf u)) undefined-function-address)
    (noniderror u 'funboundp)))

(de flambdalinkp (u)
  % is U an interpreted function?
  (if (idp u)
    (weq (symfnc (idinf u)) lambda-link-address)
    (noniderror u 'flambdalinkp)))

(de fcodep (u)
  % is U a compiled function?
  (if (idp u)
    (and  (not (funboundp u)) (not (flambdalinkp u)))
    (noniderror u 'fcodep)))

(de makefunbound (u)
  % Make U an undefined function
  (if (idp u)
    (progn (plantunbound (idinf u))
           nil)
    (noniderror u 'makefunbound)))

(de makeflambdalink (u)
  % Make U an interpreted function
  (if (idp u)
    (progn (plantlambdalink (idinf u))
           nil)
    (noniderror u 'makeflambdalink)))

(de makefcode (u codeptr)
  % Make U a compiled function
  (when (idp u)
    (if (codep codeptr)
      (progn (plantcodepointer (idinf u) (inf codeptr))
             nil)
      (noniderror u 'makefcode))))

(de makeforeigncode (u codeptr)
  % Make U a compiled function                                            
  (when (idp u)
    (if (codep codeptr)
      (progn (plantforeigncodepointer (idinf u) (inf codeptr))
             nil)
      (noniderror u 'makeforeigncode))))

(de getfcodepointer (u)
  % Get code pointer for U
  (if (idp u)
    (mkcode (symfnc (idinf u)))
    (noniderror u 'getfcodepointer)))
